// TestSuite.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "../TestScript/SimpleHeap.h"
#include <assert.h>

//////////////////////////////////////////////////////////////////////////
TEST(LuaState, creation1)
{
	LuaState* state = LuaState::Create();
	CHECK(state);
	LuaState::Destroy(state);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaStateOwner, creationDestruction)
{
	LuaStateOwner state;
}


//////////////////////////////////////////////////////////////////////////
int CountLuaObjects(LuaState* state)
{
	int count = 0;
	LuaObject* curObj = state->GetHeadObject()->m_next;
	while (curObj != state->GetTailObject())
	{
		count++;
		curObj = curObj->m_next;		
	}

	return count;
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreationBareConstructor)
{
	LuaObject obj;
	CHECK(obj.m_state == NULL);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreationWithCounting)
{
	{
		LuaStateOwner state;
		LuaObject obj(state);
		CHECK(CountLuaObjects(state) == 1);
	}

	{
		LuaStateOwner state;
		LuaObject obj(state);
		{
			LuaObject obj2(state);
			CHECK(CountLuaObjects(state) == 2);
		}
		CHECK(CountLuaObjects(state) == 1);
	}

	{
		LuaStateOwner state;
		LuaObject obj(state);
		{
			LuaObject obj2(state);
			LuaObject obj3(state);
			CHECK(CountLuaObjects(state) == 3);
		}
		CHECK(CountLuaObjects(state) == 1);
	}
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreationBareConstructorWithAssignment)
{
	LuaStateOwner state;
	LuaObject obj;
	CHECK(obj.m_state == NULL);
	obj = state->GetGlobals();
	CHECK(obj.m_state == state);
	CHECK(obj.Type() == LUA_TTABLE);
	CHECK(obj.IsTable());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreationWithLuaStatePointer)
{
	LuaStateOwner state;
	LuaObject obj(state);
	CHECK(obj.IsNil());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreationWithLuaStatePointerAndStackIndex)
{
	LuaStateOwner state;
	LuaObject obj(state, LUA_GLOBALSINDEX);
	CHECK(obj.IsTable());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CopyConstructorFromLuaStackObject)
{
	LuaStateOwner state;
	LuaStackObject stackObject = state->GetGlobals_Stack();
	LuaObject obj(stackObject);
	CHECK(obj.IsTable());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CopyConstructorFromLuaObject)
{
	LuaStateOwner state;
	LuaObject obj1(state->GetGlobals());
	LuaObject obj2(obj1);
	CHECK(state->Equal(obj1, obj2));
	CHECK(obj1.IsTable());
	CHECK(state->Equal(obj1, LuaObject(state->GetGlobals_Stack())));
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, AssignmentOperatorFromLuaObject)
{
	LuaStateOwner state;
	LuaObject obj1 = state->GetGlobals();		// Copy constructor
	LuaObject obj2;
	obj2 = obj1;								// Assignment operator
	CHECK(state->Equal(obj1, obj2));
	CHECK(obj1.IsTable());
	CHECK(state->Equal(obj1, LuaObject(state->GetGlobals_Stack())));
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, AssignmentOperatorFromLuaStackObject)
{
	LuaStateOwner state;
	LuaStackObject obj1 = state->GetGlobals_Stack();
	LuaObject obj2;
	obj2 = obj1;								// Assignment operator
	CHECK(state->Equal(LuaObject(obj1), obj2));
	CHECK(obj2.IsTable());
	CHECK(state->Equal(obj2, state->GetGlobals()));
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, Casts)
{
	LuaStateOwner state;
	LuaObject obj(state);

	// operator casts
	LuaState* var = obj;
	CHECK(var == obj.m_state);

	lua_State* var2 = obj;
	CHECK(var2 == obj.m_state->m_state);

	var = obj.GetState();
	CHECK(var == obj.m_state);

	var2 = obj.GetCState();
	CHECK(var2 == obj.m_state->m_state);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, Assign)
{
	LuaStateOwner state;
	LuaObject obj(state);
	CHECK(obj.Type() == LUA_TNIL);
	CHECK(obj.IsNil());
	CHECK(strcmp(obj.TypeName(), "nil") == 0);

	obj.AssignNil(state);
	CHECK(obj.Type() == LUA_TNIL);
	CHECK(obj.IsNil());
	CHECK(strcmp(obj.TypeName(), "nil") == 0);

	obj.AssignBoolean(state, true);
	CHECK(obj.Type() == LUA_TBOOLEAN);
	CHECK(obj.IsBoolean());
	CHECK(obj.GetBoolean() == true);
	CHECK(strcmp(obj.TypeName(), "boolean") == 0);

	obj.AssignNumber(state, 5.5);
	CHECK(obj.Type() == LUA_TNUMBER);
	CHECK(obj.IsNumber());
	CHECK(obj.GetNumber() == 5.5);
	CHECK(obj.GetInteger() == 5);		// Should downcast.
	CHECK(strcmp(obj.TypeName(), "number") == 0);

	obj.AssignString(state, "Hello");
	CHECK(obj.Type() == LUA_TSTRING);
	CHECK(obj.IsString());
	CHECK(strcmp(obj.GetString(), "Hello") == 0);
	CHECK(strcmp(obj.TypeName(), "string") == 0);

	obj.AssignWString(state, L"Hello");
	CHECK(obj.Type() == LUA_TWSTRING);
	CHECK(obj.IsWString());
	CHECK(wcscmp(obj.GetWString(), L"Hello") == 0);
	CHECK(strcmp(obj.TypeName(), "wstring") == 0);

	obj.AssignUserData(state, (void*)0x12345678);
	CHECK(obj.Type() == LUA_TUSERDATA);
	CHECK(obj.IsUserData());
	CHECK(!obj.IsLightUserData());
	CHECK(obj.GetUserData() == (void*)0x12345678);
	CHECK(strcmp(obj.TypeName(), "userdata") == 0);

	obj.AssignLightUserData(state, (void*)0x12345678);
	CHECK(obj.Type() == LUA_TLIGHTUSERDATA);
	CHECK(obj.IsUserData());
	CHECK(obj.IsLightUserData());
	CHECK(obj.GetUserData() == (void*)0x12345678);
	CHECK(strcmp(obj.TypeName(), "userdata") == 0);

	// AssignObject test
	LuaObject obj2(state);
	obj2.AssignNumber(state, 6.0);
	CHECK(obj2.IsNumber()  &&  obj2.GetNumber() == 6.0);

	obj.AssignObject(obj2);
	CHECK(obj.Type() == LUA_TNUMBER);
	CHECK(obj.IsNumber());
	CHECK(obj.GetNumber() == 6.0);
	CHECK(obj == obj2);
	CHECK(strcmp(obj.TypeName(), "number") == 0);

	// AssignTable test
	obj.AssignNewTable(state);
	CHECK(obj.Type() == LUA_TTABLE);
	CHECK(obj.IsTable());
	CHECK(strcmp(obj.TypeName(), "table") == 0);

	// Can't test AssignTObject here.
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetTable_Array)
{
	LuaStateOwner state;
	LuaObject obj;
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	for (int i = 0; i < 500; ++i)
	{
		obj.SetNumber(i, i * 2);
	}

	CHECK(obj.GetTableCount() == 500);	// This is slow, because it has to count all items.
	CHECK(obj.GetN() == 499);		// Why?  Lua is 1-based and GetN() starts counting
		// from 1 and then only for contiguous entries.

	// Remove an entry from the middle of the array.
	obj.SetNil(250);
	CHECK(obj.GetTableCount() == 499);
	CHECK(obj.GetN() == 249);		// Why is this 249?  There are actually 499
		// entries in the table, but LuaObject::GetN() only counts the
		// contiguous entries up to the first nil it finds.

	// Remove the element 250 instead.
	obj.Remove(250);
	CHECK(obj.GetTableCount() == 498);
	CHECK(obj.GetN() == 248);		// Ack, 248?  Yep.  LuaObject::Remove(),
		// which internally calls table.remove(), just decrements the current
		// N by 1.  It didn't pay attention to the fact that we just removed
		// the nil value, thereby making there be 498 contiguous entries.
		// We have to correct for it with a SetN() call.

	// Set the correct count.
	obj.SetN(498);
	CHECK(obj.GetTableCount() == 498);
	CHECK(obj.GetN() == 498);

	// Insert some elements.
	int tableCount1 = obj.GetTableCount();
	CHECK(tableCount1 == 498);
	LuaObject tempObj;
	tempObj.AssignNewTable(state);
	obj.Insert(100, tempObj);
	int tableCount2 = obj.GetTableCount();
	CHECK(tableCount1 == 498);
	obj.Insert(101, tempObj);
	CHECK(obj.GetTableCount() == 499);
	CHECK(obj.GetN() == 500);

	// Verify the inserted elements are tables.
	LuaObject verifyObj = obj[100];
	CHECK(verifyObj.IsTable());
	CHECK(verifyObj == tempObj);

	verifyObj = obj[101];
	CHECK(verifyObj.IsTable());
	CHECK(verifyObj == tempObj);

	verifyObj = obj[102];
	CHECK(!verifyObj.IsTable());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetTable_Array_With_Sort)
{
	LuaStateOwner state;
	LuaObject obj;
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	for (int i = 1; i <= 500; ++i)
	{
		obj.SetNumber(i, 500 - i * 2);
	}

	obj.SetN(500);
	obj.Sort();

	for (int i = 1; i <= 500; ++i)
	{
		LuaObject numObj = obj[i];
		CHECK(numObj.IsNumber());
		int num = numObj.GetInteger();
		CHECK(num == -(502 - i * 2));
	}
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetNil)
{
	LuaStateOwner state;
	LuaObject obj;
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// SetNil(number)
	obj.SetNumber(1, 5);
	CHECK(obj[1].GetNumber() == 5);
	obj.SetNil(1);
	CHECK(obj[1].IsNil());

	// SetNil(string)
	obj.SetNumber("Hello", 6);
	CHECK(obj["Hello"].GetNumber() == 6);
	obj.SetNil("Hello");
	CHECK(obj["Hello"].IsNil());

	// SetNil(object)
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetNumber(stringObj, 7);
	CHECK(obj[stringObj].GetNumber() == 7);
	obj.SetNil(stringObj);
	CHECK(obj[stringObj].IsNil());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetBoolean)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetBoolean(1, true);
	CHECK(obj[1].GetBoolean() == true);

	// string
	obj.SetBoolean("Hello", false);
	CHECK(obj["Hello"].GetBoolean() == false);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetBoolean(stringObj, true);
	CHECK(obj[stringObj].GetBoolean() == true);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetNumber)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetNumber(1, 5);
	CHECK(obj[1].GetNumber() == 5);

	// string
	obj.SetNumber("Hello", 6);
	CHECK(obj["Hello"].GetNumber() == 6);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetNumber(stringObj, 7);
	CHECK(obj[stringObj].GetNumber() == 7);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetString)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetString(1, "Test1");
	CHECK(strcmp(obj[1].GetString(), "Test1") == 0);

	// string
	obj.SetString("Hello", "Test2");
	CHECK(strcmp(obj["Hello"].GetString(), "Test2") == 0);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetString(stringObj, "Test3");
	CHECK(strcmp(obj[stringObj].GetString(), "Test3") == 0);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetWString)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetWString(1, L"Test1");
	CHECK(wcscmp(obj[1].GetWString(), L"Test1") == 0);

	// string
	obj.SetWString("Hello", L"Test2");
	CHECK(wcscmp(obj["Hello"].GetWString(), L"Test2") == 0);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetWString(stringObj, L"Test3");
	CHECK(wcscmp(obj[stringObj].GetWString(), L"Test3") == 0);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetUserData)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetUserData(1, (void*)0x12345678);
	CHECK(obj[1].GetUserData() == (void*)0x12345678);

	// string
	obj.SetUserData("Hello", (void*)0x87654321);
	CHECK(obj["Hello"].GetUserData() == (void*)0x87654321);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetUserData(stringObj, (void*)0x02468024);
	CHECK(obj[stringObj].GetUserData() == (void*)0x02468024);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetLightUserData)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	// number
	obj.SetLightUserData(1, (void*)0x12345678);
	CHECK(obj[1].GetUserData() == (void*)0x12345678);

	// string
	obj.SetLightUserData("Hello", (void*)0x87654321);
	CHECK(obj["Hello"].GetUserData() == (void*)0x87654321);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	obj.SetLightUserData(stringObj, (void*)0x02468024);
	CHECK(obj[stringObj].GetUserData() == (void*)0x02468024);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, SetObject)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	LuaObject testObj(state);

	// number
	testObj.AssignBoolean(state, true);
	obj.SetObject(1, testObj);
	CHECK(obj[1] == testObj);

	// string
	testObj.AssignNumber(state, 5);
	obj.SetObject("Hello", testObj);
	CHECK(obj["Hello"] == testObj);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	testObj.AssignString(state, "Stuff");
	obj.SetObject(stringObj, testObj);
	CHECK(obj[stringObj] == testObj);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CreateTable)
{
	LuaStateOwner state;
	LuaObject globalsObj = state->GetGlobals();
	CHECK(globalsObj.IsTable());

	// number
	LuaObject obj = globalsObj.CreateTable(1);
	obj.SetNumber(500, 50);
	CHECK(globalsObj[1].IsTable());
	CHECK(globalsObj[1][500].IsNumber());
	CHECK(globalsObj[1][500].GetNumber() == 50);

	// string
	obj = globalsObj.CreateTable("Hello");
	obj.SetBoolean("MyStuff", true);
	CHECK(globalsObj["Hello"].IsTable());
	CHECK(globalsObj["Hello"]["MyStuff"].IsBoolean());
	CHECK(globalsObj["Hello"]["MyStuff"].GetBoolean() == true);

	// object
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");

	obj = globalsObj.CreateTable(stringObj);
	obj.SetBoolean("OtherStuff", false);
	CHECK(globalsObj[stringObj].IsTable());
	CHECK(globalsObj[stringObj]["OtherStuff"].IsBoolean());
	CHECK(globalsObj[stringObj]["OtherStuff"].GetBoolean() == false);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, GetByFuncs)
{
	LuaStateOwner state;
	LuaObject obj(state);
	obj.AssignNewTable(state);
	CHECK(obj.IsTable());

	LuaStackObject stackObj = state->PushString("Hello");

	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");

	obj.SetNumber(1, 5);
	obj.SetNumber("Hello", 6);
	obj.SetNumber(stringObj, 7);

	CHECK(obj[1].GetNumber() == 5);
	CHECK(obj["Hello"].GetNumber() == 6);
	CHECK(obj[stackObj].GetNumber() == 6);
	CHECK(obj[stringObj].GetNumber() == 7);
	CHECK(obj.GetByIndex(1).GetNumber() == 5);
	CHECK(obj.GetByName("Hello").GetNumber() == 6);
	CHECK(obj.GetByObject(stackObj).GetNumber() == 6);
	CHECK(obj.GetByObject(stringObj).GetNumber() == 7);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, StrLen)
{
	LuaStateOwner state;
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	CHECK(stringObj.StrLen() == 4);

	stringObj.AssignWString(state, L"Wide String");
	CHECK(stringObj.StrLen() == 11);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, PushStack)
{
	LuaStateOwner state;
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	LuaStackObject stackObj = stringObj.PushStack();
	CHECK(state->GetTop() == 1);
	LuaStackObject stack2Obj = stringObj.PushStack();
	CHECK(state->GetTop() == 2);
	CHECK(stackObj == stack2Obj);
	CHECK(stackObj.IsString());
	CHECK(strcmp(stackObj.GetString(), "Test") == 0);
	state->Pop();
	CHECK(state->GetTop() == 1);
}


//////////////////////////////////////////////////////////////////////////
static int LS_AddPrint(LuaState* state)
{
	LuaStack args(state);
	assert(args[1].IsNumber());
	assert(args[2].IsNumber());
	assert(args[3].IsString());

	lua_Number add = args[1].GetNumber() + args[2].GetNumber();
	printf("%f_%s\n", add, args[3].GetString());

	return 0;
}


static int LS_Add(LuaState* state)
{
	LuaStack args(state);
	assert(args[1].IsNumber());
	assert(args[2].IsNumber());

	lua_Number add = args[1].GetNumber() + args[2].GetNumber();
	state->PushNumber(add);

	return 1;
}

class TestObject
{
public:
	TestObject(float startNumber) :
		m_startNumber(startNumber)
	{
	}

	int LS_Mul(LuaState* state)
	{
		LuaStack args(state);
		assert(args[1].IsNumber());
		assert(args[2].IsNumber());

		lua_Number value = m_startNumber + args[1].GetNumber() * args[2].GetNumber();
		state->PushNumber(value);

		return 1;
	}

protected:
	float m_startNumber;
};


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, PCall_FunctionCallNoReturnValue)
{
	LuaStateOwner state;
	state->GetGlobals().Register("AddPrint", LS_AddPrint);

	LuaObject addPrintObj = state->GetGlobals()["AddPrint"];
	CHECK(addPrintObj.IsFunction());
	CHECK(addPrintObj.IsCFunction());

	int top = state->GetTop();
	LuaCall call = addPrintObj;
	call << 5 << 10 << "Hello" << LuaRun();
	CHECK(top == state->GetTop());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, PCall_FunctionCallWithReturnValue)
{
	LuaStateOwner state;
	state->GetGlobals().Register("Add", LS_Add);

	LuaObject addObj = state->GetGlobals()["Add"];
	CHECK(addObj.IsFunction());
	CHECK(addObj.IsCFunction());

	int top = state->GetTop();
	{
		LuaAutoBlock autoBlock(state);
		LuaCall call = addObj;
		LuaStackObject resultObj = call << 5 << 10 << LuaRun();
		CHECK(resultObj.IsNumber());
		CHECK(resultObj.GetNumber() == 15);
		CHECK(top + 1 == state->GetTop());
	}

	CHECK(top == state->GetTop());

	// Now, don't accept the return value.
	{
		LuaAutoBlock autoBlock(state);
		LuaCall call = addObj;
		LuaStackObject resultObj = call << 5 << 10 << LuaRun();
		CHECK(resultObj.IsNumber());
		CHECK(top == state->GetTop() - 1);
	}
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, FunctionCallNoReturnValue)
{
	LuaStateOwner state;
	state->GetGlobals().Register("AddPrint", LS_AddPrint);

	LuaObject addPrintObj = state->GetGlobals()["AddPrint"];
	CHECK(addPrintObj.IsFunction());
	CHECK(addPrintObj.IsCFunction());

	int top = state->GetTop();
	LuaCall call = addPrintObj;
	call << 5 << 10 << "Hello" << LuaRun();
	CHECK(top == state->GetTop());
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, FunctionCallWithReturnValue)
{
	LuaStateOwner state;
	state->GetGlobals().Register("Add", LS_Add);

	LuaObject addObj = state->GetGlobals()["Add"];
	CHECK(addObj.IsFunction());
	CHECK(addObj.IsCFunction());

	int top = state->GetTop();
	{
		LuaAutoBlock autoBlock(state);
		LuaCall call = addObj;
		LuaStackObject resultObj = call << 5 << 10 << LuaRun();
		CHECK(resultObj.IsNumber());
		CHECK(resultObj.GetNumber() == 15);
		CHECK(top + 1 == state->GetTop());
	}

	CHECK(top == state->GetTop());

	// Now, don't accept the return value.
	{
		LuaAutoBlock autoBlock(state);
		LuaCall call = addObj;
		LuaStackObject resultObj = call << 5 << 10 << LuaRun(0);
		CHECK(resultObj.IsNone());
		CHECK(top == state->GetTop());
	}
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, DirectFunctionCall)
{
	LuaStateOwner state;
	state->GetGlobals().Register("Add", LS_Add);
	state->GetGlobals().Register("AddPrint", LS_AddPrint);
	
	TestObject obj(5);
	state->GetGlobals().Register("Mul", obj, &TestObject::LS_Mul);

	LuaFunction<int> addFunction = state->GetGlobals()["Add"];
	LuaFunction<void> addPrintFunction = state->GetGlobals()["AddPrint"];
	LuaFunction<float> mulFunction = state->GetGlobals()["Mul"];

	int top = state->GetTop();
	int result = addFunction(5, 10);
	CHECK(result == 15);
	int afterTop = state->GetTop();
	CHECK(top == afterTop);

	addPrintFunction(20, 30, "My String");

	top = state->GetTop();
	float fresult = mulFunction(5, 10);
	CHECK(fresult == 5 + 5 * 10);
	afterTop = state->GetTop();
	CHECK(top == afterTop);
}


//////////////////////////////////////////////////////////////////////////
int DFR_Add(int num1, int num2)
{
	return num1 + num2;
}

void DFR_AddPrint(float num1, float num2, const char* string)
{
	float add = num1 + num2;
	printf("%f_%s\n", add, string);
}

class DFRObject
{
public:
	DFRObject(float startValue) :
		m_startValue(startValue)
	{
	}

	float Mul(float num1, float num2)
	{
		return m_startValue + num1 * num2;
	}

protected:
	float m_startValue;
};

TEST(LuaObject, DirectFunctionRegister)
{
	LuaStateOwner state;
	state->GetGlobals().RegisterDirect("Add", DFR_Add);
	state->GetGlobals().RegisterDirect("AddPrint", DFR_AddPrint);

	DFRObject obj(5);
	state->GetGlobals().RegisterDirect("Mul", obj, DFRObject::Mul);

	LuaFunction<int> addFunction = state->GetGlobals()["Add"];
	LuaFunction<void> addPrintFunction = state->GetGlobals()["AddPrint"];
	LuaFunction<float> mulFunction = state->GetGlobals()["Mul"];

	int top = state->GetTop();
	int result = addFunction(5, 10);
	CHECK(result == 15);
	int afterTop = state->GetTop();
	CHECK(top == afterTop);

	addPrintFunction(20.2, 30.5, "My String");

	top = state->GetTop();
	float fresult = mulFunction(5, 10);
	CHECK(fresult == 5 + 5 * 10);
	afterTop = state->GetTop();
	CHECK(top == afterTop);
}


//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, CheckConvertibleTypes)
{
	LuaStateOwner state;
	LuaObject globalsObj = state->GetGlobals();
	globalsObj.SetNumber("Number", 501);

	LuaObject numObj = globalsObj["Number"];
	CHECK(numObj.IsNumber());
	CHECK(!numObj.IsString());
	CHECK(!numObj.IsWString());
	CHECK(numObj.IsConvertibleToNumber());
	CHECK(numObj.IsConvertibleToString());
	CHECK(numObj.IsConvertibleToWString());

	lua_Number num = numObj.ToNumber();
	CHECK(numObj.IsNumber());

	const char* str = numObj.ToString();
	CHECK(numObj.IsString());
	CHECK(strcmp(str, "501") == 0);
	CHECK(numObj.IsConvertibleToNumber());
	CHECK(!numObj.IsConvertibleToWString());

	num = numObj.ToNumber();
	CHECK(numObj.IsString());		// Once in string form, it stays there.

	globalsObj.SetNumber("Number", 501);
	numObj = globalsObj["Number"];

	const wchar_t* wstr = numObj.ToWString();
	CHECK(numObj.IsWString());
	CHECK(wcscmp(wstr, L"501") == 0);
	CHECK(numObj.IsConvertibleToNumber());
	CHECK(!numObj.IsConvertibleToString());

	num = numObj.ToNumber();
	CHECK(numObj.IsWString());		// Again, once in string form, it stays there.

	globalsObj.SetNumber("Number", 501);
	numObj = globalsObj["Number"];

	int len = numObj.ToStrLen();
	CHECK(len == 3);
	CHECK(numObj.IsString());
}

	
//////////////////////////////////////////////////////////////////////////
TEST(LuaObject, MetaTable)
{
	LuaStateOwner state;

	// Make a table for the metatable.
	LuaObject metaTableObj(state);
	metaTableObj.AssignNewTable(state);
	
	// Set up the indexing.
	LuaObject stringObj(state);
	stringObj.AssignString(state, "Test");
	metaTableObj.SetString(1, "MyLittleNumber");
	metaTableObj.SetString("__name", "MyLittleMetaTable");
	metaTableObj.SetString(stringObj, "MyLittleObject");
	
	// Make the metatable be the next lookup index.
	metaTableObj.SetObject("__index", metaTableObj);

	// Create a global called MyTable and assign the metatable.
	LuaObject myTableObj = state->GetGlobals().CreateTable("MyTable");
	myTableObj.SetMetaTable(metaTableObj);

	// Look it up again.
	LuaObject testTableObj = state->GetGlobals()["MyTable"];
	LuaObject testMetaTableObj = testTableObj.GetMetaTable();
	CHECK(metaTableObj == testMetaTableObj);
	CHECK(metaTableObj["__name"].IsString());
	CHECK(strcmp(metaTableObj["__name"].GetString(), "MyLittleMetaTable") == 0);

	// Test GetBy*() functions
	CHECK(testTableObj.GetByIndex(1).IsString());
	CHECK(strcmp(testTableObj.GetByIndex(1).GetString(), "MyLittleNumber") == 0);
	CHECK(testTableObj.GetByName("__name").IsString());
	CHECK(strcmp(testTableObj.GetByName("__name").GetString(), "MyLittleMetaTable") == 0);
	CHECK(testTableObj.GetByObject(stringObj).IsString());
	CHECK(strcmp(testTableObj.GetByObject(stringObj).GetString(), "MyLittleObject") == 0);
	CHECK(testTableObj.GetByIndex(2).IsNil());

	// Test operator functions
	CHECK(testTableObj[1].IsNil());
	CHECK(testTableObj["__name"].IsNil());
	CHECK(testTableObj[stringObj].IsNil());
}


//////////////////////////////////////////////////////////////////////////
static int CallbackFunction(LuaState* state)
{
	// For purposes of getting locals.
	return 0;
}


TEST(LuaState, Callback)
{
	LuaStateOwner state;

	state->GetGlobals().Register("Callback", CallbackFunction);
	state->DoString("Callback()");
}


//////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    TestResult tr;
    TestRegistry::runAllTests(tr);

	return 0;
}

