/**
	\file
	\author Original code by Renato Cerqueira and Vinicius Almendra.

	Modifications by Joshua Jensen.
 */

#include <string.h>

#include <ole2.h>
#include <ocidl.h>

#include <assert.h>
#include <stdio.h>
#include <math.h>

#include "LuaPlus/LuaPlus.h"

#include "luabeans.h"
#include "luacom.h"
#include "tLuaDispatch.h"
#include "tLuaCOM.h"
#include "tLuaCOMException.h"

#include "tUtil.h"


////////////////////////
//                    //
//  FUNCOES INTERNAS  //
//                    //
////////////////////////

/*
* Carrega uma Typelib associada a um ProgID
*/

static ITypeLib *loadTypeLibFromProgID(const char *pcProgID, int version_major)
{
	wchar_t* wcClsid;
	wchar_t* wcProgId;
	CLSID clsid;
	HRESULT hr;
	ITypeLib *ptlib = NULL;

	/* converte ProgID para OLESTR */
	wcProgId = (wchar_t*) malloc( (strlen(pcProgID) + 1) * sizeof(wchar_t));
	mbstowcs(wcProgId, pcProgID, strlen(pcProgID)+1);

	hr = CLSIDFromProgID(wcProgId, &clsid);

	free(wcProgId);
	wcProgId = NULL;

	if (FAILED(hr))
		return NULL;

	hr = StringFromCLSID(clsid, &wcClsid);
	if (FAILED(hr))
		return NULL;

	/* converte CLSID para string normal */
	char* pcClsid = (char*) malloc( (wcslen(wcClsid) + 1) * sizeof(char));
	wcstombs(pcClsid, wcClsid,wcslen(wcClsid)+1);

	DWORD size = 38*3; /*{F37C8063-4AD5-101B-B826-00DD01103DE1}*/
	BYTE *bLibID = (BYTE *) malloc(size);
	HKEY iid_key, obj_key, typelib_key;

	/* extrai do registry type library (GUID e versao) */

	RegOpenKeyEx(HKEY_CLASSES_ROOT,"CLSID", 0, KEY_READ, &iid_key);
	RegOpenKeyEx(iid_key, pcClsid, 0, KEY_READ, &obj_key);

	free(pcClsid);
	pcClsid = NULL;

	RegOpenKeyEx(obj_key, "TypeLib",0, KEY_READ, &typelib_key);
	RegQueryValueEx(typelib_key, NULL, NULL, NULL, bLibID, &size);

	RegCloseKey(iid_key);
	RegCloseKey(obj_key);
	RegCloseKey(typelib_key);

	wchar_t* wcTypelib= (wchar_t*) 
		malloc( (strlen((char *) bLibID) + 1) * sizeof(wchar_t));
	mbstowcs(wcTypelib, (char *) bLibID, strlen((char *) bLibID)+1);
	free(bLibID);

	CLSID libid;
	CLSIDFromString(wcTypelib, &libid);
	free(wcTypelib);

	hr = LoadRegTypeLib(libid, version_major, 0, 0, &ptlib);

	if(FAILED(hr))
		return NULL;

	return ptlib;
}

/*
*  loadTypeLibFromName
*  Carrega uma typelib a partir de um arquivo TLB
*/

static ITypeLib *loadTypeLibFromName(const char *pcFilename)
{
	HRESULT hr;
	ITypeLib *ptlib;
	wchar_t *wcFilename = NULL;

	wcFilename = (wchar_t *) malloc((strlen(pcFilename)+1)*sizeof(wchar_t));
	mbstowcs(wcFilename, pcFilename, strlen(pcFilename)+1);

	hr = LoadTypeLibEx(wcFilename, REGKIND_NONE, &ptlib);

	if(FAILED(hr))
		return NULL;

	return ptlib;
}


///////////////////////////////////
//                               //
//  FUNCOES EXPORTADAS PARA LUA  //
//                               //
///////////////////////////////////

/*
* lc_ShowHelp
*
* Parametros: 
*
*  (1) objeto luacom
*
*   Mostra Help da type library associada ao objeto
*   luacom
*/

static int lc_ShowHelp(LuaState* state)
{
	char *pHelpFile = NULL; 
	unsigned long context = 0;

	LuaStack args(state);
	tLuaCOM* luacom = (tLuaCOM*)tLuaCOM::lbeans->from_lua( args[1] );
	if(luacom == NULL)
		return 0;

	luacom->getHelpInfo(&pHelpFile, &context);

	if(pHelpFile != NULL)
	{
		if(context != 0)
			WinHelp(NULL, pHelpFile, HELP_CONTEXT, context);
		else
			WinHelp(NULL, pHelpFile, HELP_FINDER, 0);
	}

	return 0;
}

/*
* lc_Connect
*
*   Parametros (Lua):
*
*    objeto luacom
*    tabela que implementara a interface de conexao
*
*   Cria um connection point utilizando a tabela
*   passada como parametro e o conecta com o objeto
*   passado tambem como parametro
*/

static int lc_Connect(LuaState* state)
{
	LuaStack args(state);
	tLuaDispatch *client_disp = NULL;
	IID iid = CLSID_NULL;
	ITypeInfo *pTypeinfo = NULL;
	int ref = 0;

	tLuaCOM* server = (tLuaCOM*) 
		tLuaCOM::lbeans->from_lua(LuaStackObject(state, 1));

	if(!server)
	{
		state->PushNil();
		return 1;
	}

	/* obtem referencia e trava tabela lua */
	args[2].Push();
	ref = state->Ref( 1 );

	pTypeinfo = server->GetDefaultEventsInterface();

	if(pTypeinfo == NULL)
	{
		state->PushNil();
		return 1;
	}

	client_disp = new tLuaDispatch(state, pTypeinfo, ref);

	tLuaCOM* client = tLuaCOM::CreateLuaCOM(state, client_disp);

	if(client == NULL)
	{
		delete client_disp;
		state->PushNil();
		return 1;
	}

	server->addConnection(client);

	tLuaCOM::lbeans->push(client); 
	return 1;
}

/*
*  lc_ImplInterfaceFromTypelib (Lua2C)
*  Implementa uma interface descrida em uma type library
*  dada
*/

static int lc_ImplInterfaceFromTypelib(LuaState* state)
{
	LuaStack args(state);
	tLuaDispatch *iluacom = NULL;
	const char* typelib_name = args[2].GetString();
	const char* pcInterface = args[3].GetString();
	ITypeLib *ptlib;

	/* obtem referencia e trava tabela lua */
	args[1].Push();
	int ref = state->Ref( 1 );

	ptlib = loadTypeLibFromName(typelib_name);

	if(ptlib == NULL)
	{
		state->PushNil();
		return 1;
	}

	iluacom = tLuaDispatch::CreateLuaDispatch(state, ptlib, pcInterface, ref);

	ptlib->Release();
	ptlib = NULL;

	if(iluacom == NULL)
	{
		state->PushNil();
		return 1;
	}

	tLuaCOM* lcom = tLuaCOM::CreateLuaCOM(state, iluacom);

	if(!lcom)
	{
		delete iluacom;
		state->PushNil();
		return 1;
	}

	tLuaCOM::lbeans->push(lcom); 
	return 1;
}

/* 
*  lc_ImplInterface (Lua2C)
*
*  Cria uma implementacao IDispatch para um
*  objeto Lua dado um ProgID e o nome da
*  interface
*
*  In: Implementation: table,
*      ProgID: string,
*      Interface: string,
*      version_major: number
*
*  Out: LuaCOM_obj (table with tag LuaCOM)
*/

static int lc_ImplInterface(LuaState* state)
{
	LuaStack args(state);
	tLuaDispatch *iluacom = NULL;
	const char* pcProgID = args[2].GetString();
	const char* pcInterface = args[3].GetString();
	int iVersion_major = (int) args[4].GetInteger();
	ITypeLib *ptlib = NULL;

	/* obtem referencia e trava tabela lua */
	args[1].Push();
	int ref = state->Ref(1);

	/* obtem typelib */
	ptlib = loadTypeLibFromProgID(pcProgID, iVersion_major);

	if(ptlib == NULL)
	{
		state->PushNil();
		return 1;
	}

	iluacom = tLuaDispatch::CreateLuaDispatch(state, ptlib, pcInterface, ref);

	ptlib->Release();
	ptlib = NULL;

	if(iluacom == NULL)
	{
		state->PushNil();
		return 1;
	}

	tLuaCOM* lcom = tLuaCOM::CreateLuaCOM(state, iluacom);

	if(lcom == NULL)
	{
		delete iluacom;
		state->PushNil();
		return 1;
	}

	tLuaCOM::lbeans->push(lcom);
	return 1;
}

/*
*  lc_CLSIDfromProgId
*  Retorna string contendo CLSID associado a um ProgID
*/

static int lc_CLSIDfromProgId(LuaState* state)
{
	LuaStack args(state);
	const char* str = args[1].GetString();
	wchar_t* clsid_str;
	wchar_t* progId;
	CLSID clsid;
	HRESULT hr;

	progId = (wchar_t*) malloc( (strlen(str) + 1) * sizeof(wchar_t));
	mbstowcs(progId,str,strlen(str)+1);

	hr = CLSIDFromProgID(progId, &clsid);
	free(progId);
	if (FAILED(hr))
	{
//		cout << "Error looking for class " << str << endl;
		state->PushNil();
		return 1;
	}
	hr = StringFromCLSID(clsid, &clsid_str);
	if (FAILED(hr))
	{
//		cout << "Error looking for prog_id"<< endl;
		state->PushNil();
		return 1;
	}

	char* id_str = (char*) malloc( (wcslen(clsid_str) + 1) * sizeof(char));
	wcstombs(id_str,clsid_str,wcslen(clsid_str)+1);
	state->PushString(id_str);
	free(id_str);
	return 1;
}

/*
*  lc_ProgIdfromCLSID
*  Retorna string contendo o ProgID associado a um CLSID
*/

static int lc_ProgIdfromCLSID(LuaState* state)
{
	LuaStack args(state);
	const char* str = args[1].GetString();
	wchar_t* clsid_str;
	LPOLESTR progId;
	CLSID clsid;
	HRESULT hr;

	clsid_str = (wchar_t*) malloc( (strlen(str) + 1) * sizeof(wchar_t));
	mbstowcs(clsid_str,str,strlen(str)+1);

	hr = CLSIDFromString(clsid_str, &clsid);
	free(clsid_str);
	if (FAILED(hr))
	{
//		cout << "Error looking for class " << str << endl;
		state->PushNil();
		return 1;
	}
	hr = ProgIDFromCLSID(clsid, &progId);
	if (FAILED(hr))
	{
//		cout << "Error looking for prog_id"<< endl;
		state->PushNil();
		return 1;
	}

	char* id_str = (char*) malloc( (wcslen(progId) + 1) * sizeof(char));
	wcstombs(id_str,progId,wcslen(progId)+1);
	state->PushString(id_str);
	free(id_str);
	return 1;
}


static bool CLSIDFromLuaObject( CLSID& clsid, LuaStackObject& arg )
{
	if (!arg.IsString())
	{
		return false;
	}

	const char* progId = arg.GetString();

	int progIdLen = strlen( progId );
	wchar_t* olestr = new wchar_t[ progIdLen + 1 ];
	mbstowcs( olestr, progId, progIdLen + 1 );
	
	HRESULT hr = CLSIDFromProgID(olestr, &clsid);

	delete[] olestr;

	return SUCCEEDED( hr );
}


/**
	args:
		1 - progId
**/
static int lc_CreateObject( LuaState* state )
{
	LuaStack args(state);

	// Get the CLSID.
	CLSID clsid;
	if ( !CLSIDFromLuaObject( clsid, args[1] ) )
	{
		state->PushNil();
		return 1;
	}

	// Create an instance.
	LPUNKNOWN punk   = NULL;
	HRESULT hr = CoCreateInstance( clsid, NULL, CLSCTX_ALL, IID_IUnknown, (void**)&punk );
	if ( FAILED( hr ) )
	{
		state->PushNil();
		return 1;
	}

	// MFC does this, and it fixed the LuaCOM bug where Access/Excel/others couldn't be
	// accessed.
	OleRun(punk);

	// Create the object we're going to pass around.
	tLuaCOM* lcom = tLuaCOM::CreateLuaCOM(state, punk);

	// Free our reference.
	punk->Release();

	if ( !lcom )
	{
		state->PushNil();
		return 1;
	}

	// Push the resulting table onto the stack.
	tLuaCOM::lbeans->push(lcom);
	return 1;
}


/*
*  lc_GetObject
*  Cria um objeto LuaCOM associado a uma instancia
*  ja' existente do objeto COM identificado pelo 
*  ProgID dado
*/

static int lc_GetActiveObject(LuaState* state)
{
	LuaStack args(state);

	// Get the CLSID.
	CLSID clsid;
	if ( !CLSIDFromLuaObject( clsid, args[1] ) )
	{
		state->PushNil();
		return 1;
	}

	// Attach to an instance.
	LPUNKNOWN punk = NULL;
	HRESULT hr = GetActiveObject(clsid,NULL,&punk);
	if (FAILED(hr))
	{
		state->PushNil();
		return 1;
	}

	// Create the object we're going to pass around.
	tLuaCOM* lcom = tLuaCOM::CreateLuaCOM(state, punk);

	// Free our reference.
	punk->Release();

	if(lcom == NULL)
	{
		state->PushNil();
		return 1;
	}

	// Push the resulting table onto the stack.
	tLuaCOM::lbeans->push(lcom);
	return 1;
}

// Incoming:
//   1 - luacom_obj
//   2 - interface
static int lc_NewObject(LuaState* state)
{
	LuaStack args(state);
	tLuaCOM* lcom = (tLuaCOM*)tLuaCOM::lbeans->from_lua(args[1]);
	const char* pcInterface = args[2].GetString();

	tLuaCOM* newLCom = lcom->NewInterface(pcInterface);
	if(lcom == NULL)
	{
		state->PushNil();
		return 1;
	}

	tLuaCOM::lbeans->push(lcom);
	return 1;
}

/*
*  lc_addConnection
*  Associa um connection point de um objeto COM com
*  outro objeto COM
*/

static int lc_addConnection(LuaState* state)
{
	tLuaCOM* server = (tLuaCOM*) 
		tLuaCOM::lbeans->from_lua(LuaStackObject(state, 1));
	tLuaCOM* client = (tLuaCOM*) 
		tLuaCOM::lbeans->from_lua(LuaStackObject(state, 2));

	if(server == NULL || client == NULL)
	{
		state->PushNil();
		return 1;
	}

	bool result = server->addConnection(client);

	if(result == false)
		state->PushNil();

	// tudo OK
	state->PushNumber(1);
	return 1; // 2
}

/*
*  lc_releaseConnection
*  Desfaz connection point associado a um objeto LuaCOM
*/

static int lc_releaseConnection(LuaState* state)
{
	LuaStackObject table(state, 1);
	tLuaCOM* lcom = (tLuaCOM*) tLuaCOM::lbeans->from_lua(table);

	lcom->releaseConnection();
	return 0;
}

//
// lc_isMember
//
//  Informa se existe algum metodo ou propriedade com
//  o nome passado como parametro em lua
//

static int lc_isMember(LuaState* state)
{
	LuaStack args(state);
	tLuaCOM* lcom = (tLuaCOM*)tLuaCOM::lbeans->from_lua( args[ 1 ] );

	if ( !lcom  ||  !args[2].IsString() )
	{
		state->PushBoolean( false );
		return 1;
	}

	const char* member_name = args[2].GetString();
	state->PushBoolean( lcom->isMember(member_name) );
	return 1;
}


static int SafeArray_gc(LuaState* state)
{
	LuaStack args(state);
	if ( !args[1].IsUserData() )
		assert( 0 );		// How did we get here?
	VARIANTARG* pvarg = (VARIANTARG*)args[1].GetUserData();
	delete pvarg;

	return 0;
}


/**
	args:
		1 - table
**/
static int lc_SafeArray( LuaState* state )
{
	LuaStack args(state);
	LuaStackObject tableObj = args[1];

	tLuaCOMTypeHandler typeHandler(state, NULL);

	VARIANTARG* pvarg = new VARIANTARG;
	TYPEDESC typedesc;
	typedesc.vt = VT_VARIANT;
	typeHandler.safearray_lua2com(typedesc, tableObj, pvarg);

	LuaObject metaTableObj = state->GetRegistry()["comSafeArrayMetaTable"];
	if (metaTableObj.IsNil())
	{
		metaTableObj = state->GetRegistry().CreateTable("comSafeArrayMetaTable");
		metaTableObj.Register("__gc", SafeArray_gc);
	}

	LuaObject outObj = state->NewUserDataBox(pvarg);
	outObj.SetMetaTable(metaTableObj);

	return 1;
}


/*
* Checa se o parametro self foi passado corretamente
*/ 

static bool check_self(LuaStackObject self)
{
	// garante existencia e garante que o parametro e' uma tabela
	// ou um userdata

	if(self.IsNone()  ||  self.IsNil()  || 
		!( self.IsTable() || self.IsUserData() )
		)
	{
		return false;
	}

	return true;
}



/////////////////////////////////////
//                                 //
//  FUNCOES EXPORTADAS PARA C/C++  //
//                                 //
/////////////////////////////////////

/*
*  luacom_idispatch2luacom
*  Recebe um ponteiro void* que deve ser
*  um ponteiro para uma implementacao de uma
*  interface IDispatch. E' criado um objeto
*  LuaCOM cuja implementacao e' dada por essa
*  interface. O objeto criado e' colocado na
*  pilha de Lua
*/

void luacom_IUnknown2LuaCOM(LuaState* state, void *punk_arg)
{
	unsigned int ninfo = 0;
	IUnknown *punk = (IUnknown *) punk_arg;

	assert(punk);
	if(punk == NULL)
	{
		state->PushNil();
		return;
	}

	tLuaCOM* lcom = tLuaCOM::CreateLuaCOM(state, punk);

	if(lcom == NULL)
	{
		state->PushNil();
		return;
	}

	tLuaCOM::lbeans->push(lcom);
}


extern "C" __declspec(dllexport) void LuaPlus_RegisterModule( LuaState* state )
{
	LuaAutoBlock autoBlock( state );

	if (::OleInitialize(NULL) != 0)
		return;

	LuaObject comObj = state->GetGlobals().CreateTable( "com" );
	comObj.Register("CreateObject", lc_CreateObject);
	comObj.Register("GetActiveObject", lc_GetActiveObject);
	comObj.Register("NewObject", lc_NewObject);
	comObj.Register("CLSIDfromProgId",lc_CLSIDfromProgId);
	comObj.Register("ImplInterface",lc_ImplInterface);
	comObj.Register("ImplInterfaceFromTypelib",lc_ImplInterfaceFromTypelib);
	comObj.Register("AddConnection", lc_addConnection);
	comObj.Register("ReleaseConnection", lc_releaseConnection);
	comObj.Register("ProgIdfromCLSID",lc_ProgIdfromCLSID);
	comObj.Register("IsMember",lc_isMember);
	comObj.Register("Connect",lc_Connect);
	comObj.Register("ShowHelp",lc_ShowHelp);
	comObj.Register("SafeArray",lc_SafeArray);

	tLuaCOM::lbeans = LuaBeans::createBeans(state, tLuaCOM::TAG);
	tLuaCOM::lbeans->reg_tagmethods();
	tLuaCOM::lbeans->reg_ud_tagmethods();
}


extern "C" __declspec(dllexport) void LuaPlus_UnregisterModule( LuaState* state )
{
	OleUninitialize();
	state->GetGlobals().SetNil( "com" );
}
